from collections import OrderedDict

import pandas as pd
from rest_framework.exceptions import APIException
from xlrd import XLRDError


class RowResult(object):
    IMPORT_TYPE_UPDATE = 'update'
    IMPORT_TYPE_NEW = 'new'
    IMPORT_TYPE_DELETE = 'delete'
    IMPORT_TYPE_SKIP = 'skip'
    IMPORT_TYPE_ERROR = 'error'

    def __init__(self):
        self.errors = []
        self.import_type = None


class Result(object):

    def __init__(self):
        self.totals = OrderedDict([(RowResult.IMPORT_TYPE_NEW, 0),
                                   (RowResult.IMPORT_TYPE_UPDATE, 0),
                                   (RowResult.IMPORT_TYPE_DELETE, 0),
                                   (RowResult.IMPORT_TYPE_SKIP, 0),
                                   (RowResult.IMPORT_TYPE_ERROR, 0)])
        self.total_rows = 0
        self.error_detail = set()

    def increment_row_result_total(self, row_result):
        if row_result.import_type:
            self.totals[row_result.import_type] += 1


class BaseResource:

    @classmethod
    def get_result_class(cls):
        """
        Returns the class used to store the result of an import.
        """
        return Result

    @classmethod
    def get_row_result_class(cls):
        """
        Returns the class used to store the result of a row import.
        """
        return RowResult

    def import_data(self, raw_data, ignore_exception=True):
        self.before_import_script()
        result = self.get_result_class()()
        for row in raw_data:
            row_result = self.import_row(row, result, ignore_exception)
            result.increment_row_result_total(row_result)
        self.after_import_script()
        return result

    def import_row(self, row, result, ignore_exception=True):
        row_result = self.get_row_result_class()()
        try:
            serializer = self.Meta.serializer_class(data=row, context=self.get_context())
            serializer.is_valid(raise_exception=True)
            validated_row = serializer.validated_data
            instance = self.get_instance(validated_row)
            import_type = self.get_import_type(instance, validated_row)
            row_result = self.dispatch(instance, import_type, serializer, row_result)
        except Exception as e:
            if ignore_exception:
                row_result.import_type = RowResult.IMPORT_TYPE_ERROR
                result.error_detail.add(str(e) + '\n')
            else:
                raise
        return row_result

    def get_raw_df_data_from_file(self, file_obj):
        """
        根据传入的 file_obj 获取 dataframe
        :param file_obj: 文件对象，可以为 excel, zip 等。
        :return: 返回组织好的 dataframe
        """
        try:
            df = pd.read_excel(io=file_obj,
                               sheet_name=self.Meta.sheet_name,
                               converters=self.Meta.head_data_types,
                               na_values=self.Meta.null_value)
        # 处理没找到对应 sheet_name 的情况。
        except XLRDError as e:
            raise APIException(str(e), code=422)
        # 将 null 值替换为 None，方便 ORM 插入
        return df.where(df.notnull(), None)

    def get_import_type(self, instance, row, default_action=RowResult.IMPORT_TYPE_SKIP):
        """
        决定此条数据的 action，分为 'create', 'update', 'ignore', 'delete'
        :param instance: 根据 import_id_fields 的数据获取的对象。
        :param row: 清洗过的原始数据
        :param default_action: 默认的 action default
        :return: 返回需要执行的action RowResult 中的 IMPORT_TYPE_NEW
        """
        raise NotImplementedError('`get_dispatch_action()` must be implemented.')

    def get_instance(self, row):
        try:
            params = {}
            for key in self.Meta.import_id_fields:
                params[key] = row.get(key)
            return self.Meta.model.objects.all().get(**params)
        except self.Meta.model.DoesNotExist:
            return None

    @staticmethod
    def dispatch(instance, import_type, serializer, row_result):
        if import_type == RowResult.IMPORT_TYPE_NEW:
            serializer.save()
            row_result.import_type = RowResult.IMPORT_TYPE_NEW
        elif import_type == RowResult.IMPORT_TYPE_UPDATE:
            serializer.instance = instance
            serializer.save()
            row_result.import_type = RowResult.IMPORT_TYPE_UPDATE
        elif import_type == RowResult.IMPORT_TYPE_SKIP:
            row_result.import_type = RowResult.IMPORT_TYPE_SKIP
        elif import_type == RowResult.IMPORT_TYPE_DELETE:
            instance.delete()
            row_result.import_type = RowResult.IMPORT_TYPE_DELETE
        else:
            # todo raise 无法识别的 import_type Error
            pass
        return row_result

    def get_context(self):
        return self.Meta.context

    def after_import_script(self):
        pass

    def before_import_script(self):
        pass

    class Meta:
        """
        导入excel时，需要的元信息。
        """
        # serializer 导入时需要的主 serializer。
        serializer_class = None
        # 用于获取 instance
        import_id_fields = None
        # 用于获取 instance 的 model
        model = None

        # 导入 excel 的目标 sheets，为 pandas read_excel sheet_name 参数
        sheet_name = 0
        # 用于将某些字符串在 read_excel 时转换成 NaN。注意只能替换字符串，不能替换其他类型，例如 数字 0
        # pandas read_excel na_value 参数 例如['-']
        null_value = None
        # 指定某些列的数据类型，防止本来是字符串类型被转换为整数类型等情况
        # sample: {column_head_1: np.str, column_head_2: np.int}
        head_data_types = None
        # 序列化时需要的其他数据，例如一些不在数据库中的数据，json等。
        context = {}
