from rest_framework.serializers import Serializer


def col_range(max_col_num, start_col=None):
    """
    确定 column_range 范围
    :param max_col_num: 最大 column_range
    :param start_col: 起始位置 'A' or 20
    :return: 'A:I'
    """
    if start_col is None:
        start_col = 1

    if isinstance(start_col, int) and start_col < max_col_num:
        start_col_label = num_to_col_label(start_col)
    elif isinstance(start_col, str) and col_label_to_num(start_col) < max_col_num:
        start_col_label = start_col
    else:
        start_col_label = 'A'

    end_col_label = num_to_col_label(max_col_num)
    return "{start}:{end}".format(start=start_col_label, end=end_col_label)


def num_to_col_label(num):
    """
    28 => 'AB'
    :param num: 非零整数
    :return: 返回整数对应的 Excel column label
    """
    letters = ''
    while num:
        mod = (num - 1) % 26
        letters += chr(mod + 65)
        num = (num - 1) // 26
    return ''.join(reversed(letters))


def col_label_to_num(label):
    """
    'AB' => 28
    :param label: Excel column label 例如 A
    :return: 返回对应的数字
    """
    label = label.upper()
    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    maps = {}
    for i in range(26):
        maps[letters[i]] = i + 1

    col = 0
    for l in label:
        col = 26 * col + maps[l]

    return col


def filter_none(obj):
    """Remove ``None`` values from tuples, lists or dictionaries. Return other objects as-is.

    :param obj: the object
    :return: collection with ``None`` values removed
    """
    if obj is None:
        return None
    new_obj = None
    if isinstance(obj, dict):
        new_obj = type(obj)((k, v) for k, v in obj.items() if k is not None and v is not None)
    if isinstance(obj, (list, tuple)):
        new_obj = type(obj)(v for v in obj if v is not None)
    if new_obj is not None and len(new_obj) != len(obj):
        return new_obj  # pragma: no cover
    return obj


def flatten_serializer(serializer_fields, field_label=None, parent_name=None):
    """
    扁平化 serializer 中定义的 nested_seriliazer
    :param serializer_fields:
    :param field_label: {field1：label1, field2:label2, ...}

    :return: {field1：label1, field2:label2, ...} and {label1:field1, label2:field2, ...}
    """

    if field_label is None:
        field_label = {}

    if parent_name is None:
        parent_name = ''

    for field_name, field_value in serializer_fields.items():

        if parent_name:
            field_name = parent_name + '.' + field_name

        if isinstance(field_value, Serializer):
            flatten_serializer(field_value._declared_fields, field_label=field_label, parent_name=field_name)
        else:
            field_label.update({field_name: field_value.label})

    return field_label


def rebuild_nested_data(source, seprator='.'):
    output = {}

    for k, v in source.items():
        current = output
        pieces = k.split(seprator)
        # iterate from the beginning until the second to last place
        for piece in pieces[:-1]:
            if piece not in current:
                # if a dict doesn't exist at an index, then create one
                current[piece] = {}
            # as you walk into the structure, update your current location
            current = current[piece]
        # The reason you're using the second to last is because the last place
        # represents the place you're actually storing the item
        current[pieces[-1]] = v

    return output
