from string import digits

import numpy
from rest_framework.exceptions import ValidationError

from pyexternal_import.resource import BaseResource, RowResult
from pyexternal_import.serializer import ValuationSheetBasicInfoSerializer
from pyexternal_import.utils import filter_none, flatten_serializer


def get_date_from_valuation_sheet(meta_info_lst):
    """
    获取估值日期
    :param meta_info_lst: 前几行基础信息
    :return:
    """
    for info_element in meta_info_lst:
        if '日期' in info_element:
            # 删除空格
            try:
                raw_date = info_element.split('：')[1]
                raw_date = raw_date.replace(' ', '')
            except IndexError:
                raw_date = info_element.split(':')[1]
                raw_date = raw_date.replace(' ', '')
            # '2019年12月31日'，'2019-1-31', '20190131'
            if '年' in raw_date:
                ret_date = raw_date.replace('年', '-').replace('月', '-').replace('日', '')
            elif '-' in raw_date:
                ret_date = raw_date
            else:
                if len(raw_date) != 8:
                    raise ValidationError('日期提取失败，20190131 模式不存在。')
                ret_date = raw_date[:4] + '-' + raw_date[4:6] + '-' + raw_date[6:8]
            return ret_date
    raise ValidationError('日期提取失败')


def get_meta_info_list(raw_dataframe, info_max_row=None):
    """
    提取 meta 信息
    :param raw_dataframe: 原始 dataframe
    :param info_max_row: meta 信息存在在 [0:info_max_row]
    :return:
    """
    if info_max_row is None:
        info_max_row = 2
    df_meta = raw_dataframe.iloc[0:info_max_row]
    df_meta = df_meta.where(df_meta.notnull(), None)
    df_meta_list = df_meta.values.flatten()
    return filter_none(df_meta_list.tolist())


def get_fund_nav(meta_info_lst):
    """
    获取单位净值
    :param meta_info_lst: 前几行的基础信息
    :return:
    """
    for info_element in meta_info_lst:
        if '单位净值' in info_element:
            nav_chips = info_element.split('.')
            # 1 + .0233
            return nav_chips[0][-1] + '.' + ''.join(c for c in nav_chips[-1] if c in digits)
    raise ValidationError('净值提取失败')


def get_meta_info(raw_df):
    """
    由于 小M 估值表 meta_info 的位置不稳定需要多次尝试 header index
    :param raw_df: 原始 dataframe
    :return:
    """
    date = None
    master_fund_nav = None
    for header_index in [1, 2, 3]:
        try:
            # 提取日期
            meta_info_list = get_meta_info_list(raw_df, info_max_row=header_index)
            date = get_date_from_valuation_sheet(meta_info_list)
            # 提取单位净值
            master_fund_nav = get_fund_nav(meta_info_list)
            break
        except (ValidationError, IndexError):
            continue
    if master_fund_nav is None or date is None:
        raise ValidationError('meta_info 提取失败，检查估值表结构。')
    return [header_index, master_fund_nav, date]


def strip_column_and_row_names(raw_dataframe, col_index=None, row_index=None):
    """
    一般第三行为 column name, 第一列为 row name，清洗 行列名。
    :param raw_dataframe: excel 原始数据
    :param col_index: column_name 所在 row index
    :param row_index: row_name 所在 col index
    :return:
    """
    # 去除 row 全为 Nan 的 column
    raw_dataframe = raw_dataframe.dropna(axis=1, how='all')
    if col_index is None:
        col_index = 2
    if row_index is None:
        row_index = 0
    # todo https://www.dataquest.io/blog/settingwithcopywarning/
    raw_dataframe.iloc[col_index] = raw_dataframe.iloc[col_index].apply(
        lambda x: x.replace(' ', '').replace('%', '').replace('(', '').replace(')', '') if isinstance(x, str) else x)
    raw_dataframe.iloc[:, row_index] = raw_dataframe.iloc[:, row_index].apply(
        lambda x: x.replace(':', '').replace('：', '').replace('-', '').replace('——', '') if isinstance(x, str) else x)
    return raw_dataframe


def get_value_from_valuation_sheets(df, loc1, row_name, loc2, alt_row=None):
    """
    loc1  ------ loc2
     -
     -
     -
     -
    row_name(alt_row) --- value
    :param df: 设置好 column name 的 pandas dataframe
    :param loc1: 如图
    :param row_name: 如图
    :param loc2: 如图
    :param alt_row: 如图
    :return: 如图 value
    """
    value = df[df[loc1] == row_name][loc2].get_values()

    if value.size == 0 and alt_row is not None:
        value = df[df[loc1] == alt_row][loc2].get_values()

    if value.size == 0:
        return None
    else:
        # 假设有重复项的时候，取第一个。
        if isinstance(value[0], str):
            # '2,007.45' => '2007.45'
            ret = value[0].replace(',', '').replace('，', '')
        else:
            ret = value[0]
        return ret


def get_position_from_valuation_sheets(df, loc_pair, ignore_list):
    """
    get rows between loc_pair[0] and loc_pair[1]
    loc_pair[0]
    .
    .
    .
    rows
    .
    .
    .
    loc_pair[1]
    :param df: 设置好 column name 的 pandas dataframe
    :param loc_pair: 如图
    :param ignore_list: 不希望存在的 row
    :return: 如图 rows
    """
    # 确定需要关注的 rows 区间，例如 ['股票投资', '债券投资']
    boolean_index = (df['科目名称'] == loc_pair[0]) | (df['科目名称'] == loc_pair[1])
    # 将区间外的所有 False 都换成 True
    first_true_index = boolean_index[boolean_index == True].index[0]
    first_row = boolean_index.index[0]
    last_true_index = boolean_index[boolean_index == True].index[-1]
    last_row = boolean_index.index[-1]
    boolean_index[first_row:first_true_index] = True
    boolean_index[last_true_index:last_row + 1] = True

    # 删除 ignore list 中非关注的 rows
    if ignore_list:
        for item in ignore_list:
            boolean_index = boolean_index | (df['科目名称'] == item)
    # 假设停牌信息不为空的时候我基础标的。logical nor
    boolean_index = boolean_index | df['停牌信息'].isnull()
    # boolean_index 开始不是从 0 index 开始，可能有 offset
    first_true_index = boolean_index[boolean_index == True].index[0] - boolean_index.index[0]
    last_true_index = boolean_index[boolean_index == True].index[-1] - boolean_index.index[0] + 1
    boolean_index[first_true_index:last_true_index] = numpy.logical_not(boolean_index[first_true_index:last_true_index])
    return df[boolean_index].reset_index(drop=True)


def get_body_from_raw_dataframe(raw_dataframe, header_index):
    # 假设估值表 body 第一行，第二列是 银行存款。
    for offset in [1, 2, 3, 4]:
        df_body = raw_dataframe.iloc[header_index + offset:]
        if df_body.iloc[0][1] == '银行存款':
            return df_body
    raise Exception('获取估值表 body 失败。')


def get_predefined_locations_value(value_location, value_alias, df_body, label_field=None):
    """
    根据 value_location, value_alias 提取对应位置的值，并根据转换表替换 key
    :param value_location:
    :param value_alias:
    :param df_body:
    :param label_field:
    :return:
    """
    basic_info = {}
    for name, location in value_location.items():
        for unify_entry, alias in value_alias.items():
            if name in alias:
                value = get_value_from_valuation_sheets(df_body, location[0], name, location[1])
                if value is not None:
                    if label_field:
                        en_name = label_field[unify_entry]
                        basic_info[en_name] = value
                    else:
                        basic_info[unify_entry] = value
                    break
    return basic_info


def get_possible_locations_value(possible_location, value_alias, df_body, label_field=None):
    """
    根据 possible_location, value_alias 提取对应位置的值，并根据转换表替换 key
    :param possible_location:
    :param value_alias:
    :param df_body:
    :param label_field:
    :return:
    """
    basic_info = {}
    for unify_entry, location in possible_location.items():
        possible_choice = value_alias[unify_entry]
        for possible_entry in possible_choice:
            value = get_value_from_valuation_sheets(df_body, location[0], possible_entry, location[1])
            if value is not None:
                if label_field:
                    en_name = label_field[unify_entry]
                    basic_info[en_name] = value
                else:
                    basic_info[unify_entry] = value
                break
    return basic_info


class ValuationSheetResource(BaseResource):

    def get_info_from_valuation_excel(self, file_obj, **kwargs):
        """
        从 complex excel 中获取清洗好的 raw_data
        :param file_obj: excel 文件
        :return: 清洗好的 raw_data
        """
        df = self.get_raw_df_data_from_file(file_obj)
        header_index, fund_nav, date = get_meta_info(df)
        cn_key_flag = kwargs.get('cn_key')
        if cn_key_flag:
            info = {'单位净值': fund_nav, '日期': date}
        else:
            info = {'fund_nav': fund_nav, 'nav_date': date}
        # 清洗行列名
        raw_dataframe = strip_column_and_row_names(df, col_index=header_index)
        # 获取 银行存款，单位净值，资产净值，实收资本，基金投资
        df_body = get_body_from_raw_dataframe(raw_dataframe, header_index)
        df_body.columns = raw_dataframe.iloc[header_index]
        df_body = self.unify_column_header(df_body)
        # 清洗 position_df 的数据
        df_body = self.validate_df_body(df_body)
        # 获取 basic_info 例如银行存款，资产净值等。
        basic_info = self.extract_basic_info(df_body, cn_key=cn_key_flag)
        info = {**info, **basic_info}
        # 获取持仓
        df_body = df_body.reset_index(drop=True)
        if self.Meta.position_loc_and_ignore_lst:
            position_dfs = dict()
            for sheet_name_prefix, locator_dict in self.Meta.position_loc_and_ignore_lst.items():
                position_df = get_position_from_valuation_sheets(df_body, locator_dict['locator_pair'],
                                                                 ignore_list=locator_dict.get('ignore_list'))
                position_dfs[sheet_name_prefix] = position_df
            info = {**info, 'position_dfs': position_dfs}
        return info

    def unify_column_header(self, df_body):
        """
        unify column_header for serialize
        :param df_body:
        :return:
        """
        rename_dict = dict()
        for name in df_body.columns:
            for unify_name, alias in self.Meta.header_alias.items():
                if name in alias:
                    rename_dict[name] = unify_name
        if rename_dict:
            df_body.rename(columns=rename_dict, inplace=True)
        return df_body

    def validate_df_body(self, df_body):
        """
        df_body 的数据清洗
        :param df_body:
        :return:
        """
        # todo round 到正确数值，方便 dfs_to_excel 精确控制表头宽度。
        # 例如
        # 确保是 numeric 的数字形式
        # raw_dataframe[['数量', '单位成本', '市值-本币']] = raw_dataframe[['数量', '单位成本', '市值-本币']].astype(np.float64)
        return df_body

    def rebuild_column_headers_for_serialize(self, df_body):
        """
        根据 serializer 定义的
        :param df_body:
        :return:
        """
        field_label = flatten_serializer(self.Meta.serializer_class._declared_fields)
        label_field = {v: k for k, v in field_label.items()}
        rename_dict = {name: label_field[name] for name in df_body.columns if name in label_field}
        if rename_dict:
            df_body.rename(columns=rename_dict, inplace=True)
        return df_body

    def extract_basic_info(self, df_body, cn_key=None):
        """
        根据 value_location 提取基础值，例如 '基金资产净值': ('科目代码', '市值'),
        :param df_body:
        :param cn_key: cn_key_flag 如果为 True 返回
        :return:
        """
        label_field = None
        if not cn_key:
            field_label = flatten_serializer(self.Meta.serializer_class._declared_fields)
            label_field = {v: k for k, v in field_label.items()}
        if self.Meta.value_location is None:
            basic_info = get_possible_locations_value(self.Meta.possible_location, self.Meta.value_alias, df_body,
                                                      label_field=label_field)
        else:
            basic_info = get_predefined_locations_value(self.Meta.value_location, self.Meta.value_alias, df_body,
                                                        label_field=label_field)
        return basic_info

    def get_import_type(self, instance, row, default_action=RowResult.IMPORT_TYPE_SKIP):
        if instance is None:
            action = RowResult.IMPORT_TYPE_NEW
        elif isinstance(instance, self.Meta.model):
            action = RowResult.IMPORT_TYPE_UPDATE
        else:
            action = default_action
        return action

    class Meta(BaseResource.Meta):
        # 导入 excel 的目标 sheets，为 pandas read_excel sheet_name 参数
        serializer_class = ValuationSheetBasicInfoSerializer
        # 导入 excel 的时候有数据的那一个 sheet
        sheet_name = 'Sheet1'
        # 用于指定哪些列有值，可以用于估值表对比。
        value_columns = ['数量', '单位成本', '成本', '行情', '市值', '估值增值']
        # position_loc_and_ignore_lst 可能出现多组数据，每组数据用一个名字来命名sheet。
        # {
        #     '上交所': {
        #         'ignore_list': [],
        #         'locator_pair': ['上交所_已上市_信用_股票_成本', '深交所_已上市_信用_股票']
        #     },
        #     '深交所': {
        #         'ignore_list': ['ETF基金', 'ETF基金成本', 'ETF基金估值增值'],
        #         'locator_pair': ['深交所_已上市_信用_股票_成本', '深交所_已上市_信用_股票_创业板']
        #     },
        #     '深交所_创业板': {
        #         'ignore_list': [],
        #         'locator_pair': ['深交所_已上市_信用_股票_创业板_成本', '应付托管费']
        #     }
        # }
        position_loc_and_ignore_lst = None
        # 为 pandas read_excel na_value 参数
        null_value = [0, '-']
        # '银行存款': ('科目名称', '市值') 表示
        """
        科目名称 ------ 市值
         -             -
         -             -
         -             -
         -             -
        银行存款 ------ value
        """
        value_alias = {
            '银行存款': {'银行存款'},
            '累计单位净值': {'累计单位净值'},
            '资产净值': {'基金资产净值', '信托资产净值', '资产净值', '资产资产净值'},
            '实收资本': {'实收资本金额', '实收信托', '实收资本'},
            '理财': {'基金理财产品成本', '理财投资', '理财产品投资', '基金投资'}
        }
        header_alias = {
            '成本占比': {'成本占净值', '成本占比'},
            '行情': {'行情', '市价'},
            '市值': {'市值-本币', '市值'},
            '估值增值': {'估值增值-本币', '估值增值'},
            '市值占比': {'市值占比', '市值占净值'},
            '成本': {'成本本币', '成本'}
        }
        possible_location = {
            '银行存款': ('科目名称', '市值'),
            '累计单位净值': ('科目代码', '科目名称'),
            '资产净值': ('科目代码', '市值'),
            '实收资本': ('科目代码', '数量'),
            '理财': ('科目名称', '市值')
        }
        # value location sample value:
        # {
        #     '银行存款': ('科目名称', '市值'),
        #     '累计单位净值': ('科目代码', '科目名称'),
        #     '基金资产净值': ('科目代码', '市值'),
        #     '实收资本金额': ('科目代码', '数量')
        # }
        value_location = None
